@@ -1,5 +1,5 @@
 /*
- *  Copyright (C) 2019-2020 Scoopta
+ *  Copyright (C) 2019-2022 Scoopta
  *  This file is part of Wofi
  *  Wofi is free software: you can redistribute it and/or modify
     it under the terms of the GNU General Public License as published by
@@ -75,7 +75,7 @@ static char* get_search_text(char* file) {
 
 static bool populate_widget(char* file, char* action, struct widget_builder* builder) {
 	GDesktopAppInfo* info = g_desktop_app_info_new_from_filename(file);
-	if(info == NULL || g_desktop_app_info_get_is_hidden(info) || g_desktop_app_info_get_nodisplay(info)) {
+	if(info == NULL || !g_app_info_should_show(G_APP_INFO(info))) {
 		return false;
 	}
 	const char* name;
@@ -119,11 +119,17 @@ static bool populate_widget(char* file, char* action, struct widget_builder* bui
 			pixbuf = gtk_icon_info_load_icon(info, NULL);
 		}
 
+		if(pixbuf == NULL) {
+			goto img_fail;
+		}
+
 		pixbuf = utils_g_resize_pixbuf(pixbuf, wofi_get_image_size() * wofi_get_window_scale(), GDK_INTERP_BILINEAR);
 
 		wofi_widget_builder_insert_image(builder, pixbuf, "icon", NULL);
 		g_object_unref(pixbuf);
 	}
+
+	img_fail:
 	wofi_widget_builder_insert_text(builder, name, "name", NULL);
 	wofi_widget_builder_insert_text(builder, generic_name, "generic-name", NULL);
 	free(generic_name);
@@ -180,6 +186,9 @@ static struct widget_builder* populate_actions(char* file, size_t* text_count) {
 
 static char* get_id(char* path) {
 	char* applications = strstr(path, "applications/");
+	if(applications == NULL) {
+		return NULL;
+	}
 	char* name = strchr(applications, '/') + 1;
 	char* id = strdup(name);
 
@@ -192,6 +201,10 @@ static char* get_id(char* path) {
 
 static struct widget* create_widget(char* full_path) {
 	char* id = get_id(full_path);
+	if(id == NULL) {
+		free(full_path);
+		return NULL;
+	}
 
 	if(map_contains(entries, id)) {
 		free(id);
@@ -231,6 +244,10 @@ static void insert_dir(char* app_dir) {
 		}
 		char* full_path = utils_concat(3, app_dir, "/", entry->d_name);
 		char* id = get_id(full_path);
+		if(id == NULL) {
+			free(full_path);
+			continue;
+		}
 
 		struct stat info;
 		stat(full_path, &info);
@@ -395,16 +412,47 @@ static void set_dri_prime(GDesktopAppInfo* info) {
 	}
 }
 
+static bool uses_dbus(GDesktopAppInfo* info) {
+	return g_desktop_app_info_get_boolean(info, "DBusActivatable");
+}
+
+static char* get_cmd(GAppInfo* info) {
+	const char* cmd = g_app_info_get_commandline(info);
+	size_t cmd_size = strlen(cmd);
+	char* new_cmd = calloc(1, cmd_size + 1);
+	size_t new_cmd_count = 0;
+	for(size_t count = 0; count < cmd_size; ++count) {
+		if(cmd[count] == '%') {
+			if(cmd[++count] == '%') {
+				new_cmd[new_cmd_count++] = cmd[count];
+			}
+		} else {
+			new_cmd[new_cmd_count++] = cmd[count];
+		}
+	}
+	if(new_cmd[--new_cmd_count] == ' ') {
+		new_cmd[new_cmd_count] = 0;
+	}
+	return new_cmd;
+}
+
 void wofi_drun_exec(const gchar* cmd) {
 	GDesktopAppInfo* info = g_desktop_app_info_new_from_filename(cmd);
 	if(G_IS_DESKTOP_APP_INFO(info)) {
 		wofi_write_cache(mode, cmd);
 		if(print_command) {
-			printf("%s\n", g_app_info_get_commandline(G_APP_INFO(info)));
+			char* cmd = get_cmd(G_APP_INFO(info));
+			printf("%s\n", cmd);
+			free(cmd);
 			exit(0);
 		} else {
 			set_dri_prime(info);
-			g_app_info_launch_uris_async(G_APP_INFO(info), NULL, NULL, NULL, launch_done, (gchar*) cmd);
+			if(uses_dbus(info)) {
+				g_app_info_launch_uris_async(G_APP_INFO(info), NULL, NULL, NULL, launch_done, (gchar*) cmd);
+			} else {
+				g_app_info_launch_uris(G_APP_INFO(info), NULL, NULL, NULL);
+				exit(0);
+			}
 		}
 	} else if(strrchr(cmd, ' ') != NULL) {
 		char* space = strrchr(cmd, ' ');
@@ -413,12 +461,13 @@ void wofi_drun_exec(const gchar* cmd) {
 		info = g_desktop_app_info_new_from_filename(cmd);
 		char* action = space + 1;
 		if(print_command) {
-			printf("%s\n", g_app_info_get_commandline(G_APP_INFO(info)));
+			char* cmd = get_cmd(G_APP_INFO(info));
+			printf("%s\n", cmd);
+			free(cmd);
 			fprintf(stderr, "Printing the command line for an action is not supported\n");
 		} else {
 			set_dri_prime(info);
 			g_desktop_app_info_launch_action(info, action, NULL);
-			utils_sleep_millis(500);
 		}
 		exit(0);
 	} else {
