@@ -1,5 +1,5 @@
 /*
- *  Copyright (C) 2019-2020 Scoopta
+ *  Copyright (C) 2019-2022 Scoopta
  *  This file is part of Wofi
  *  Wofi is free software: you can redistribute it and/or modify
     it under the terms of the GNU General Public License as published by
@@ -26,8 +26,10 @@
 #include <stdint.h>
 
 #include <sys/stat.h>
+#include <sys/wait.h>
 
 #include <utils.h>
+#include <match.h>
 #include <config.h>
 #include <utils_g.h>
 #include <property_box.h>
@@ -39,12 +41,9 @@
 #include <pango/pango.h>
 #include <gdk/gdkwayland.h>
 
-static const char* terminals[] = {"kitty", "termite", "alacritty", "gnome-terminal", "weston-terminal"};
+#define PROTO_VERSION(v1, v2) (v1 < v2 ? v1 : v2)
 
-enum matching_mode {
-	MATCHING_MODE_CONTAINS,
-	MATCHING_MODE_FUZZY
-};
+static const char* terminals[] = {"kitty", "termite", "alacritty", "foot", "gnome-terminal", "weston-terminal"};
 
 enum location {
 	LOCATION_CENTER,
@@ -103,6 +102,10 @@ static uint32_t line_count = 0;
 static bool dynamic_lines;
 static struct wl_list mode_list;
 static pthread_t mode_thread;
+static bool has_joined_mode = false;
+static char* copy_exec = NULL;
+static char* pre_display_cmd = NULL;
+static bool single_click = false;
 
 static struct map* keys;
 
@@ -129,13 +132,13 @@ static void nop() {}
 static void add_interface(void* data, struct wl_registry* registry, uint32_t name, const char* interface, uint32_t version) {
 	(void) data;
 	if(strcmp(interface, zwlr_layer_shell_v1_interface.name) == 0) {
-		shell = wl_registry_bind(registry, name, &zwlr_layer_shell_v1_interface, version);
+		shell = wl_registry_bind(registry, name, &zwlr_layer_shell_v1_interface, PROTO_VERSION(version, 4));
 	} else if(strcmp(interface, wl_output_interface.name) == 0) {
 		struct output_node* node = malloc(sizeof(struct output_node));
-		node->output = wl_registry_bind(registry, name, &wl_output_interface, version);
+		node->output = wl_registry_bind(registry, name, &wl_output_interface, PROTO_VERSION(version, 4));
 		wl_list_insert(&outputs, &node->link);
 	} else if(strcmp(interface, zxdg_output_manager_v1_interface.name) == 0) {
-		output_manager = wl_registry_bind(registry, name, &zxdg_output_manager_v1_interface, version);
+		output_manager = wl_registry_bind(registry, name, &zxdg_output_manager_v1_interface, PROTO_VERSION(version, 3));
 	}
 }
 
@@ -393,6 +396,39 @@ static GtkWidget* create_label(char* mode, char* text, char* search_text, char*
 
 	setup_label(mode, WOFI_PROPERTY_BOX(box));
 
+	char* nodetext = text;
+	if (pre_display_cmd != NULL) {
+		text = NULL;
+		FILE *fp_labeltext;
+		char *cmd_labeltext;
+		char line[128]; // you'd think this caps the line's length to 128, but it's just a buffer which due to the nature of fgets() splits on lines
+		size_t size = 0;
+		// first, prepare cmd_labeltext to be each entry's actual comamand to run, aka replacing 'cat %s' to be 'cat filename'
+		if ((asprintf(&cmd_labeltext, pre_display_cmd, nodetext)) == -1) {
+			fprintf(stderr, "error parsing pre_display_cmd to run\n");
+			exit(EXIT_FAILURE);
+		}
+		// then, run the command
+		fp_labeltext = popen(cmd_labeltext, "r");
+		if (fp_labeltext == NULL) {
+			fprintf(stderr, "error executing '%s'\n", cmd_labeltext);
+			exit(EXIT_FAILURE);
+		} else if (fgets(line, sizeof(line), fp_labeltext) != NULL) {
+			// lastly, read the output of said command, and put it into the text variable to be used for the label widgets
+			// consider using 'printf %.10s as your --pre-display-cmd to limit a string to a determined width. 10 here is an example
+			size += strlen(line)+1; // we need place for the \0 of strcpy
+			text = (char *) realloc(text, size);
+			strcpy(text, line);
+			while (fgets(line, sizeof(line), fp_labeltext) != NULL) {
+				size += strlen(line);
+				text = (char *) realloc(text, size);
+				strncat(text, line, size);
+			}
+		}
+
+		pclose(fp_labeltext);
+	}
+
 	if(allow_images) {
 		parse_images(WOFI_PROPERTY_BOX(box), text, true);
 	} else {
@@ -475,11 +511,26 @@ static void update_surface_size(void) {
 	}
 
 	gtk_window_resize(GTK_WINDOW(window), width, height);
-	gtk_widget_set_size_request(scroll, width, height);
+	GtkAllocation alloc;
+	gtk_widget_get_allocated_size(entry, &alloc, NULL);
+	if(outer_orientation == GTK_ORIENTATION_HORIZONTAL) {
+		if(alloc.width > 0) {
+			gtk_widget_set_size_request(scroll, width - alloc.width, height);
+		}
+	} else {
+		if(alloc.height > 0) {
+			gtk_widget_set_size_request(scroll, width, height - alloc.height);
+		}
+	}
 }
 
 static void widget_allocate(GtkWidget* widget, GdkRectangle* allocation, gpointer data) {
 	(void) data;
+	if(widget == entry) {
+		update_surface_size();
+		return;
+	}
+
 	if(resize_expander) {
 		return;
 	}
@@ -533,7 +584,7 @@ static gboolean _insert_widget(gpointer data) {
 		gtk_expander_set_label_widget(GTK_EXPANDER(parent), box);
 
 		GtkWidget* exp_box = gtk_list_box_new();
-		gtk_list_box_set_activate_on_single_click(GTK_LIST_BOX(exp_box), FALSE);
+		gtk_list_box_set_activate_on_single_click(GTK_LIST_BOX(exp_box), single_click);
 		g_signal_connect(exp_box, "row-activated", G_CALLBACK(activate_item), NULL);
 		gtk_container_add(GTK_CONTAINER(parent), exp_box);
 		for(size_t count = 1; count < node->action_count; ++count) {
@@ -599,7 +650,10 @@ static gboolean _insert_widget(gpointer data) {
 }
 
 static gboolean insert_all_widgets(gpointer data) {
-	pthread_join(mode_thread, NULL);
+	if(!has_joined_mode) {
+		pthread_join(mode_thread, NULL);
+		has_joined_mode = true;
+	}
 	struct wl_list* modes = data;
 	if(modes->prev == modes) {
 		return FALSE;
@@ -918,26 +972,16 @@ static gboolean filter_proxy(GtkFlowBoxChild* row) {
 	if(GTK_IS_EXPANDER(box)) {
 		box = gtk_expander_get_label_widget(GTK_EXPANDER(box));
 	}
-	const gchar* text = wofi_property_box_get_property(WOFI_PROPERTY_BOX(box), "filter");
-	if(filter == NULL || strcmp(filter, "") == 0) {
-		return TRUE;
-	}
-	if(text == NULL) {
-		return FALSE;
-	}
-	if(insensitive) {
-		return strcasestr(text, filter) != NULL;
-	} else {
-		return strstr(text, filter) != NULL;
-	}
+	const gchar* text =
+			wofi_property_box_get_property(WOFI_PROPERTY_BOX(box), "filter");
+	return match_for_matching_mode(filter, text, matching, insensitive);
 }
 
-static gboolean do_filter(GtkFlowBoxChild* row, gpointer data) {
-	(void) data;
-	gboolean ret = filter_proxy(row);
+static void do_resize_surface_after_filter(GtkFlowBoxChild *row, gboolean filter_return) {
 
-	if(gtk_widget_get_visible(GTK_WIDGET(row)) == !ret && dynamic_lines) {
-		if(ret) {
+	if(gtk_widget_get_visible(GTK_WIDGET(row)) == !filter_return &&
+			dynamic_lines) {
+		if(filter_return) {
 			++line_count;
 		} else {
 			--line_count;
@@ -952,82 +996,18 @@ static gboolean do_filter(GtkFlowBoxChild* row, gpointer data) {
 		}
 	}
 
-	gtk_widget_set_visible(GTK_WIDGET(row), ret);
+	gtk_widget_set_visible(GTK_WIDGET(row), filter_return);
+}
+
+static gboolean do_filter(GtkFlowBoxChild* row, gpointer data) {
+	(void) data;
+	gboolean ret = filter_proxy(row);
+
+	do_resize_surface_after_filter(row, ret);
 
 	return ret;
 }
 
-static gint fuzzy_sort(const gchar* text1, const gchar* text2) {
-	char* _filter = strdup(filter);
-	size_t len = strlen(_filter);
-
-	char* t1 = strdup(text1);
-	size_t t1l = strlen(t1);
-
-	char* t2 = strdup(text2);
-	size_t t2l = strlen(t2);
-
-	if(insensitive) {
-		for(size_t count = 0; count < len; ++count) {
-			char chr = _filter[count];
-			if(isalpha(chr)) {
-				_filter[count] = tolower(chr);
-			}
-		}
-		for(size_t count = 0; count < t1l; ++count) {
-			char chr = t1[count];
-			if(isalpha(chr)) {
-				t1[count] = tolower(chr);
-			}
-		}
-		for(size_t count = 0; count < t2l; ++count) {
-			char chr = t2[count];
-			if(isalpha(chr)) {
-				t2[count] = tolower(chr);
-			}
-		}
-	}
-
-	size_t dist1 = utils_distance(t1, _filter);
-	size_t dist2 = utils_distance(t2, _filter);
-	free(_filter);
-	free(t1);
-	free(t2);
-	return dist1 - dist2;
-}
-
-static gint contains_sort(const gchar* text1, const gchar* text2) {
-	char* str1, *str2;
-
-	if(insensitive) {
-		str1 = strcasestr(text1, filter);
-		str2 = strcasestr(text2, filter);
-	} else {
-		str1 = strstr(text1, filter);
-		str2 = strstr(text2, filter);
-	}
-	bool tx1 = str1 == text1;
-	bool tx2 = str2 == text2;
-	bool txc1 = str1 != NULL;
-	bool txc2 = str2 != NULL;
-
-	if(tx1 && tx2) {
-		return 0;
-	} else if(tx1) {
-		return -1;
-	} else if(tx2) {
-		return 1;
-	} else if(txc1 && txc2) {
-		return 0;
-	} else if(txc1) {
-		return -1;
-	} else if(txc2) {
-		return 1;
-	} else {
-		return 0;
-	}
-}
-
 static gint do_sort(GtkFlowBoxChild* child1, GtkFlowBoxChild* child2, gpointer data) {
 	(void) data;
 	gtk_flow_box_get_child_at_index(GTK_FLOW_BOX(inner_box), 0);
@@ -1050,7 +1030,7 @@ static gint do_sort(GtkFlowBoxChild* child1, GtkFlowBoxChild* child2, gpointer d
 		return index1 - index2;
 	}
 
-	uint64_t fallback = 0;
+	int fallback = 0;
 	switch(sort_order) {
 	case SORT_ORDER_DEFAULT:
 		fallback = index1 - index2;
@@ -1067,24 +1047,7 @@ static gint do_sort(GtkFlowBoxChild* child1, GtkFlowBoxChild* child2, gpointer d
 	if(filter == NULL || strcmp(filter, "") == 0) {
 		return fallback;
 	}
-
-	gint primary = 0;
-	switch(matching) {
-	case MATCHING_MODE_CONTAINS:
-		primary = contains_sort(text1, text2);
-		if(primary == 0) {
-			return fallback;
-		}
-		return primary;
-	case MATCHING_MODE_FUZZY:
-		primary = fuzzy_sort(text1, text2);
-		if(primary == 0) {
-			return fallback;
-		}
-		return primary;
-	default:
-		return 0;
-	}
+	return sort_for_matching_mode(text1, text2, fallback, matching, filter, insensitive);
 }
 
 static void select_first(void) {
@@ -1193,6 +1156,50 @@ static void do_hide_search(void) {
 	update_surface_size();
 }
 
+static void do_copy(void) {
+	GList* children = gtk_flow_box_get_selected_children(GTK_FLOW_BOX(inner_box));
+	if(children->data != NULL && gtk_widget_has_focus(children->data)) {
+		GtkWidget* widget = gtk_bin_get_child(children->data);
+		if(GTK_IS_EXPANDER(widget)) {
+			GtkWidget* box = gtk_bin_get_child(GTK_BIN(widget));
+			GtkListBoxRow* row = gtk_list_box_get_selected_row(GTK_LIST_BOX(box));
+			if(row == NULL) {
+				widget = gtk_expander_get_label_widget(GTK_EXPANDER(widget));
+			} else {
+				widget = gtk_bin_get_child(GTK_BIN(row));
+			}
+		}
+		if(WOFI_IS_PROPERTY_BOX(widget)) {
+			const gchar* action = wofi_property_box_get_property(WOFI_PROPERTY_BOX(widget), "action");
+			if(action == NULL) {
+				return;
+			}
+
+			int fds[2];
+			if (pipe(fds) == -1) {
+				perror("pipe broken");
+				exit(EXIT_FAILURE);
+			}
+			if(fork() == 0) {
+				close(fds[1]);
+				dup2(fds[0], STDIN_FILENO);
+				execlp(copy_exec, copy_exec, NULL);
+				fprintf(stderr, "%s could not be executed: %s\n", copy_exec, strerror(errno));
+				exit(errno);
+			}
+			close(fds[0]);
+
+			if(write(fds[1], action, strlen(action)) <= 0) {
+				fprintf(stderr, "fd pipe failed to write\n");
+			}
+
+			close(fds[1]);
+
+			while(waitpid(-1, NULL, WNOHANG) > 0);
+		}
+	}
+}
+
 static bool do_key_action(GdkEvent* event, char* mod, void (*action)(void)) {
 	if(mod != NULL) {
 		GdkModifierType mask = get_mask_from_name(mod);
@@ -1340,12 +1347,14 @@ static gboolean focus_entry(GtkWidget* widget, GdkEvent* event, gpointer data) {
 	(void) data;
 	if(widget == entry && dbus != NULL) {
 		GError* err = NULL;
-		g_dbus_proxy_call_sync(dbus, "SetVisible", g_variant_new("(b)", event->focus_change.in), G_DBUS_CALL_FLAGS_NONE, 2000, NULL, &err);
+		GVariant* ret = g_dbus_proxy_call_sync(dbus, "SetVisible", g_variant_new("(b)", event->focus_change.in), G_DBUS_CALL_FLAGS_NONE, 2000, NULL, &err);
 		if(err != NULL) {
 			if(err->code != G_DBUS_ERROR_SERVICE_UNKNOWN) {
 				fprintf(stderr, "Error while changing OSK state %s\n", err->message);
 			}
 			g_error_free(err);
+		} else {
+			g_variant_unref(ret);
 		}
 	}
 	return FALSE;
@@ -1593,7 +1602,7 @@ void wofi_init(struct map* _config) {
 	char* password_char = map_get(config, "password_char");
 	exec_search = strcmp(config_get(config, "exec_search", "false"), "true") == 0;
 	bool hide_scroll = strcmp(config_get(config, "hide_scroll", "false"), "true") == 0;
-	matching = config_get_mnemonic(config, "matching", "contains", 2, "contains", "fuzzy");
+	matching = config_get_mnemonic(config, "matching", "contains", 3, "contains", "multi-contains", "fuzzy");
 	insensitive = strcmp(config_get(config, "insensitive", "false"), "true") == 0;
 	parse_search = strcmp(config_get(config, "parse_search", "false"), "true") == 0;
 	location = config_get_mnemonic(config, "location", "center", 18,
@@ -1611,6 +1620,9 @@ void wofi_init(struct map* _config) {
 	dynamic_lines = strcmp(config_get(config, "dynamic_lines", "false"), "true") == 0;
 	char* monitor = map_get(config, "monitor");
 	char* layer = config_get(config, "layer", "top");
+	copy_exec = config_get(config, "copy_exec", "wl-copy");
+	pre_display_cmd = map_get(config, "pre_display_cmd");
+	single_click = strcmp(config_get(config, "single_click", "false"), "true") == 0;
 
 	keys = map_init_void();
 
@@ -1628,6 +1640,7 @@ void wofi_init(struct map* _config) {
 	char* key_pgdn = config_get(config, "key_pgdn", "Page_Down");
 	char* key_expand = config_get(config, "key_expand", "");
 	char* key_hide_search = config_get(config, "key_hide_search", "");
+	char* key_copy = config_get(config, "key_copy", "Control_L-c");
 
 	parse_mods(key_up, move_up);
 	parse_mods(key_down, move_down);
@@ -1641,6 +1654,7 @@ void wofi_init(struct map* _config) {
 	parse_mods(key_pgdn, move_pgdn);
 	parse_mods(key_expand, do_expand);
 	parse_mods(key_hide_search, do_hide_search);
+	parse_mods(key_copy, do_copy);
 
 	modes = map_init_void();
 
@@ -1657,9 +1671,8 @@ void wofi_init(struct map* _config) {
 	gtk_window_set_decorated(GTK_WINDOW(window), FALSE);
 
 	if(!normal_window) {
-		GdkDisplay* disp = gdk_display_get_default();
 		wl_list_init(&outputs);
-		wl = gdk_wayland_display_get_wl_display(disp);
+		wl = gdk_wayland_display_get_wl_display(gdk_display_get_default());
 
 		if(wl == NULL) {
 			fprintf(stderr, "Failed to connect to wayland compositor\n");
@@ -1749,6 +1762,7 @@ void wofi_init(struct map* _config) {
 	gtk_widget_set_name(outer_box, "outer-box");
 	gtk_container_add(GTK_CONTAINER(window), outer_box);
 	entry = gtk_search_entry_new();
+	g_signal_connect(entry, "size-allocate", G_CALLBACK(widget_allocate), NULL);
 
 	gtk_widget_set_name(entry, "input");
 	gtk_entry_set_placeholder_text(GTK_ENTRY(entry), prompt);
@@ -1767,7 +1781,6 @@ void wofi_init(struct map* _config) {
 	scroll = gtk_scrolled_window_new(NULL, NULL);
 	gtk_widget_set_name(scroll, "scroll");
 	gtk_container_add(GTK_CONTAINER(outer_box), scroll);
-	gtk_widget_set_size_request(scroll, width, height);
 	if(hide_scroll) {
 		gtk_scrolled_window_set_policy(GTK_SCROLLED_WINDOW(scroll), GTK_POLICY_EXTERNAL, GTK_POLICY_EXTERNAL);
 	}
@@ -1780,22 +1793,15 @@ void wofi_init(struct map* _config) {
 	gtk_widget_set_valign(inner_box, valign);
 
 	gtk_widget_set_name(inner_box, "inner-box");
-	gtk_flow_box_set_activate_on_single_click(GTK_FLOW_BOX(inner_box), FALSE);
+	gtk_flow_box_set_activate_on_single_click(GTK_FLOW_BOX(inner_box), single_click);
 
 	GtkWidget* wrapper_box = gtk_box_new(GTK_ORIENTATION_VERTICAL, 0);
 	gtk_box_set_homogeneous(GTK_BOX(wrapper_box), TRUE);
 	gtk_container_add(GTK_CONTAINER(wrapper_box), inner_box);
 	gtk_container_add(GTK_CONTAINER(scroll), wrapper_box);
 
-	switch(matching) {
-	case MATCHING_MODE_CONTAINS:
-		gtk_flow_box_set_filter_func(GTK_FLOW_BOX(inner_box), do_filter, NULL, NULL);
-		gtk_flow_box_set_sort_func(GTK_FLOW_BOX(inner_box), do_sort, NULL, NULL);
-		break;
-	case MATCHING_MODE_FUZZY:
-		gtk_flow_box_set_sort_func(GTK_FLOW_BOX(inner_box), do_sort, NULL, NULL);
-		break;
-	}
+	gtk_flow_box_set_filter_func(GTK_FLOW_BOX(inner_box), do_filter, NULL, NULL);
+	gtk_flow_box_set_sort_func(GTK_FLOW_BOX(inner_box), do_sort, NULL, NULL);
 
 	g_signal_connect(inner_box, "child-activated", G_CALLBACK(activate_item), NULL);
 	g_signal_connect(inner_box, "selected-children-changed", G_CALLBACK(select_item), NULL);
